#!/usr/bin/python

import random
import getopt
import sys
import re
import os
import subprocess
import string
import time
from math import ceil

##--- class
class History :

  def __init__(self,hf,m,hl):
    self.historyFile = hf
    self.maxTimeToConsider = m
    self.resetHalflife = hl   
    self.inStartup = False
    self.error = 0
    self.currentMcq = 0
    self.previousMcq = 0
    self.beginStartupTime = 0
    self.totalArea = 0

  def storeError(self,e):
    self.error = e
    secs  = time.time()
    f = open(self.historyFile, 'a')
    value = ('seconds: ' + str(secs) + ': error: ' + str(self.error) + '\n')
    f.write(value)
    f.close()

  def storeMcq(self,m):
    self.currentMcq = m
    secs  = time.time()
    f = open(self.historyFile, 'a')
    value = ('seconds: ' + str(secs) + ': mcq: ' + str(self.currentMcq) + '\n')
    f.write(value)
    f.close()

  def beginStartup(self):
    secs  = time.time()
    f = open(self.historyFile, 'a')
    value = ('seconds: ' + str(secs) + ': beginStartupEvent' + '\n')
    f.write(value)
    f.close()
    self.inStartup = True

  def endStartup(self):
    secs  = time.time()
    f = open(self.historyFile, 'a')
    value = ('seconds: ' + str(secs) + ': endStartupEvent' + '\n')
    f.write(value)
    f.close()
    self.inStartup = False

  def analyseHistoryFile(self):
    self.totalArea = 0
    topTime = 0
    prevTime = -1
    prevErr  = -1
    self.inStartup = False
    self.beginStartupTime = 0
    self.previousMcq = 0

    f = open(self.historyFile, 'r')
    errorLineCount = 0
    errorLines = []
    line = f.readline()
    while(line != ''):
      errorMatch = re.match('^.*error.*$',line)
      if errorMatch is not None:
        errorLineCount = errorLineCount + 1
        errorLines.append(line)
      else:
        eventMatch = re.match('^seconds: ([0-9\.]+): beginStartupEvent',line)
        if eventMatch is not None:
          self.beginStartupTime = int(float(eventMatch.group(1)))
          self.inStartup = True
        else:
          eventMatch = re.match('^seconds: ([0-9\.]+): endStartupEvent',line)
          if eventMatch is not None:
            self.inStartup = False
          else:
            mcqMatch = re.match('^seconds: ([0-9\.]+): mcq: ([0-9]+)$',line)
            if mcqMatch is not None:
              self.previousMcq = int(float(mcqMatch.group(2))) * 8

      line = f.readline()
    f.close()

    if errorLineCount <= 1:
      print 'DEBUG totalArea: ',self.totalArea
      return

    # Old lookback area calc. Obsolete.
    if (self.maxTimeToConsider >= 0):
      errorLines.reverse()
      errorLineCount = 0
      for line in errorLines:
        errorLineCount = errorLineCount + 1
        parts = line.split(': ')
        t = int(float(parts[1])) * -1
        e = int(float(parts[3]))
  
        if (errorLineCount == 1):
          topTime = t
          prevTime = t
          prevErr = e
        else:
          if abs(topTime - t ) <= self.maxTimeToConsider:
            deltaT = t - prevTime
            approxE = (e + prevErr) / 2
            area = (deltaT * approxE)
            self.totalArea = self.totalArea + area
            prevTime = t
            prevErr = e
    else:
      # Work out totalArea using decay
      timeNow = time.time()
      errorLineCount = 0
      for line in errorLines:
        errorLineCount = errorLineCount + 1
        parts = line.split(': ')
        t = int(float(parts[1])) 
        e = int(float(parts[3]))
  
        if (errorLineCount == 1):
          topTime = t
          prevTime = t
          prevErr = e
        else:   
          deltaT = t - prevTime
          approxE = (e + prevErr) / 2
          area = (deltaT * approxE)
          hl = self.resetHalflife
          timeExpired = timeNow - t + ((t - prevTime)/2.0)

          decayedArea = area * pow(1.0/2.0,timeExpired/hl)
          self.totalArea = self.totalArea + decayedArea
          prevTime = t
          prevErr = e

    print 'DEBUG totalArea: ',self.totalArea
    return 

##--- class
class Job :
  def __init__(self,g,r,j):
    self.globalJobId = g
    self.requestCpus = r
    self.jobStatus = j
      
##--- class
class Queue :

  def getTotalQueued(self):
    return self.totalQueued 
  def getTotalSingleQueued(self):
    return self.totalSingleQueued 
  def getTotalMultiQueued(self):
    return self.totalMultiQueued 
  def getTotalRunning(self):
    return self.totalRunning 
  def getTotalSingleJobsRunning(self):
    return self.totalSingleJobsRunning 
  def getTotalMultiJobsRunning(self):
    return self.totalMultiJobsRunning 
  def getTotalHeld(self):
    return self.totalHeld 
  def getTotalDontKnow(self):
    return self.totalDontKnow 
  def getTotalSingleUnicoresRunning(self):
    return self.totalSingleJobsRunning * 1
  def getTotalMultiUnicoresRunning(self):
    return self.totalMultiJobsRunning * 8

  def __init__(self):
    self.jobs = {}
    self.totalQueued = 0
    self.totalSingleQueued = 0
    self.totalMultiQueued = 0
    self.totalRunning = 0
    self.totalSingleJobsRunning = 0
    self.totalMultiJobsRunning = 0
    self.totalHeld = 0
    self.totalDontKnow = 0

    worked,code,report,err = runCommand ("condor_q -long ")
    if (worked != True):
      print 'ERROR in Queue class, condor_q command did not run'
      sys.exit(1)
    if (code != 0):
      print 'ERROR in Queue class, condor_q command gave return code ', code, ', stderr was ', err
      sys.exit(1)
      
    lines = string.split(report, '\n')

    globalJobId = -1
    requestCpus = -1
    jobStatus = -1

    for l in lines:
      matchObj = re.match('^$',l)  
      if matchObj is not None:
        job = Job(globalJobId,requestCpus,jobStatus)
        self.jobs[globalJobId] = job
        if (jobStatus == 1):
          self.totalQueued = self.totalQueued + 1
          if (requestCpus == 1):
            self.totalSingleQueued = self.totalSingleQueued + 1
          else:
            self.totalMultiQueued = self.totalMultiQueued + 1
        else:
          if (jobStatus == 2):
            self.totalRunning = self.totalRunning + 1
            if (requestCpus == 1):
              self.totalSingleJobsRunning = self.totalSingleJobsRunning + 1
            else:
              self.totalMultiJobsRunning = self.totalMultiJobsRunning + 1 

          else:
            if (jobStatus == 5):
              self.totalHeld = self.totalHeld + 1
            else:
              self.totalDontKnow = self.totalDontKnow + 1
        
      else:
        halves = l.split(' = ')
        if halves[0] == 'GlobalJobId':
          globalJobId = halves[1]
        if halves[0] == 'RequestCpus':
          requestCpus = int(halves[1])
        if halves[0] == 'JobStatus':
          jobStatus = int(halves[1])

##--- class
class Node :

  """ Holds a Node """

  # Constructor, with initial attribute values
  def __init__(self,name):
    self.name = name
    self.detectedCpus = 0    # Don't know
    self.totalSlotCpus = 0   # How much is available on this node
    self.totalUsed     = 0   # How much is used 
    self.slack         = 0   # How much is not used
    self.slots = []
    self.current_slot = {}
    self.hasSlotWith8 = False

  def __cmp__(self, other):
    if self.totalSlotCpus < other.totalSlotCpus:
      return 1
    elif self.totalSlotCpus > other.totalSlotCpus:
      return -1
    else:
      return 0

  def getHasSlotWith8 (self):
    return self.hasSlotWith8

  def getName (self):
    return self.name

  def getSlot(self,s):
    max = len(self.slots) - 1
    if (s < 0 or s > max):
      return None
    slot = self.slots[s]
    return slot

  def getState(self):
    slot = self.slots[0]
    return slot['State']

  def readNodeState (self):
    worked,code,report,err = runCommand ("condor_status -long " + self.name)
    if (worked != True):
      print 'WARNING in Node class, condor_status command did not run for ',self.name
      return False
    if (code != 0):
      print 'WARNING in Node class, for node', self.name, ', condor_status command gave return code ', code, ', stderr was ', err
      return False
    lines = string.split(report, '\n')

    slot_count = 0

    for l in lines:
      matchObj = re.match('^$',l)  
      if matchObj is not None:
        self.slots.append(self.current_slot)
        self.current_slot = {}
        slot_count = slot_count + 1
      else:
        halves = l.split(' = ')
        self.current_slot[halves[0]] = halves[1]
        if self.detectedCpus == 0 and halves[0] == 'DetectedCpus':
          self.detectedCpus = int(halves[1])
        if self.totalSlotCpus == 0 and halves[0] == 'TotalSlotCpus':
          self.totalSlotCpus = int(halves[1])
        if halves[0] == 'Cpus':   # When in first slot, how much is slack. When in later slots, how much is used.
          if int(halves[1]) >= 8:
            self.hasSlotWith8 = True

          if slot_count == 0:
            self.slack = self.slack + int(halves[1])
          else:
            self.totalUsed = self.totalUsed + int(halves[1])  
    return True

#--- sub
def runCommand(cmd):
  worked = False
  statusCode = -99
  stdout_value = ''
  stderr_value = ''
  command=cmd.split(' ',1)
  try:
    p = subprocess.Popen([cmd ], stdout=subprocess.PIPE, shell=True)
    stdout_value, stderr_value = p.communicate()
    worked=True
    statusCode=p.returncode
  except Exception :
    worked=False
  return worked,statusCode,stdout_value, stderr_value

#--- sub
def listNodes():
  worked,code,report,err = runCommand ("condor_status -format '%s\n' Machine")
  if (worked != True):
    print 'ERROR in listNodes, condor_status command did not run'
    sys.exit(1)
  if (code != 0):
    print 'ERROR in listNodes, condor_status command gave return code ', code, ', stderr was ', err
    sys.exit(1)

  lines = string.split(report[0:-1], '\n')

  nodeNames = []
  for l in lines:
    nodeNames.append(l)

  nodeset = set(nodeNames)
  nodeNames = list(nodeset)
  nodeNames.sort()
  return nodeNames

#--- sub
def selectNodesToDrain(nl,num):
  if num <= 0:
    return []
  count = 0
  #random.shuffle(nl)
  nl = sorted(nl)
  listToDrain = []
  for n in nl:
    state = n.getState()
    if 'Drained' not in state :
      if not n.getHasSlotWith8() :
        listToDrain.append(n)
        count = count + 1
        if (count >= num):
          return listToDrain
  return listToDrain
  
#--- sub
def startDraining(nl,test):
  if (len(nl) == 0):
    print 'Nothing to do'
  for n in nl:
    print 'Will drain ',n.getName()
    if not test: 
      worked,code,report,err = runCommand ("condor_drain " + n.getName())
      if (worked != True):
        print 'WARNING in startDraining, condor_drain command did not run for ',n.getName()
        return 
      if (code != 0):
        print 'WARNING in startDraining, for node', n.getName(), ', condor_drain command gave return code ', code, ', stderr was ', err
        return 

#--- sub
def cancelSet(nl,test):
  for n in nl:
    print 'In cancelSet, will cancel drain of ',n.getName()
    if not test: 
      worked,code,report,err = runCommand ("condor_drain  -cancel " + n.getName())
      if (worked != True):
        print 'WARNING in cancelSet, condor_drain -cancel command did not run for ',n.getName()
        return 
      if (code != 0):
        print 'WARNING in cancelSet, for node', n.getName(), ', condor_drain -cancel command gave return code ', code, ', stderr was ', err
        return 

#--- sub
def putReadyNodesInService(nl,test):
  stillDraining = []
  for n in nl:
    resumed = False
    i = 0
    s = n.getSlot(i)
    while (s ):
      k = 'Cpus'
      cpus = s[k]
      if (int(cpus) >= 8):
        print 'In putReadyNodesInService, will cancel drain of ',n.getName()
        if not test: 
          resumed = True
          worked,code,report,err = runCommand ("condor_drain  -cancel " + n.getName())
          if (worked != True):
            print 'WARNING in putReadyNodesInService, condor_drain -cancel command did not run for ',n.getName()
            resumed = False
          if (code != 0):
            print 'WARNING in putReadyNodesInService, for node', n.getName(), ', condor_drain -cancel command gave return code ', code, ', stderr was ', err
            resumed = False

      i = i + 1
      s = n.getSlot(i)
    if (not resumed):
      stillDraining.append(n)
  return stillDraining

#--- sub
def getDrainingNodes(nl):
  draining = []
  for n in nl:
    state = n.getState()
    if 'Drained' in state :
      draining.append(n)
  return draining

#--- sub
def usage():
  print 'This program controls the drain rate on a condor server '
  print 'using a process controller.'
  print ''
  print 'The options to this program are:'
  print ' -s  --setpoint      250     the setpoint value'
  print ' -p  --propband      200     proportional band'
  print ' -r  --reset         10000   reset time'
  print ' -i  --resethalflife 86400.0 reset halflife'
  print ' -l  --lookback      86400   look back time (obs.)'
  print ' -m  --maxtodrain    9       max that can drain at once'
  print ' -t  --test                  test mode '
  print ' -k  --keepgoing             keep going, don\'t cancel draining'
  print ' -c  --coef          1.5     Ramp up coefficient'
  print ' -h  --halflife      3600.0  Ramp up halflife'
  print ''

#--- sub
def initOptions(o):

  """ Sets the defaults and read the options """

  # Set defaults (not used)
  o['setpoint']      = 250
  o['propband']      = 200
  o['resethalflife'] = 86400.0
  o['lookback']      = 86400
  o['reset']         = 10000
  o['maxtodrain']    = 10000
  o['test']      = False
  o['keepgoing'] = False
  o['coef'] = 1.5
  o['halflife'] = 3600.0

  # Read the options
  try:
    options, remainder = getopt.getopt(sys.argv[1:], 's:p:r:l:m:tkc:h:i:', 
      [ 'setpoint=', 'propband=','reset=','lookback=','maxtodrain=','test','keepgoing','coef=','halflife=','resethalflife='])

  except getopt.GetoptError:
    usage(); sys.exit(1)

  if len(options) <= 1:
    usage(); sys.exit(1)

  # Store the options
  hadLookback = False
  hadResethalflife= False
  for opt, arg in options:
    try:
      if opt in ('-t', '--test'):
        o['test'] = True
      if opt in ('-k', '--keepgoing'):
        o['keepgoing'] = True
      if opt in ('-s', '--setpoint'):
        o['setpoint'] =  int(arg)
      elif opt in ('-p', '--propband'):
        o['propband'] = int(arg)
      elif opt in ('-l', '--lookback'):
        o['lookback'] = int(arg)
        o['resethalflife'] = -1
        hadLookback = True
      elif opt in ('-r', '--reset'):
        o['reset'] = int(arg)
      elif opt in ('-m', '--maxtodrain'):
        o['maxtodrain'] = int(arg)
      elif opt in ('-c', '--coef'):
        o['coef'] = float(arg)
      elif opt in ('-h', '--halflife'):
        o['halflife'] = float(arg)
      elif opt in ('-i', '--resethalflife'):
        o['resethalflife'] = float(arg)
        o['lookback'] = -1
        hadResethalflife= True
    except ValueError:
      usage(); sys.exit(1)

  if (hadLookback and hadResethalflife):
    print 'Can not use both lookback and resethalflife.\n'
    usage(); sys.exit(1)
    
#--- main

options = {}
initOptions(options)
proportionalBand = options['propband']
resetTime = options['reset']
maxToDrain = options['maxtodrain']
test = options['test']

q = Queue()

history = History('/tmp/drainBoss.txt',options['lookback'],options['resethalflife'])

nodeList = []
for n in listNodes():
  node = Node(n)
  if (node.readNodeState()):
    nodeList.append(node)

if (len(nodeList) == 0):
  print 'ERROR in main, could find no nodes at all'
  sys.exit(1)

sp = options['setpoint'] 
upperBound = int(sp + proportionalBand / 2)
lowerBound = int(sp - proportionalBand / 2)

# Set the total controller output to be a float
totCo = 0.0

# Get some general parameters; current value, gain and absolute error
cvm = q.getTotalMultiUnicoresRunning()
gain = 1 / float((upperBound - lowerBound))
ae = sp - cvm

# Write out the absolute error
history.storeError(ae)

# Calculate the proportional component of the controller output
propCo_unbounded = ae * float(gain )

# Calculate the integral component of the controller output
history.analyseHistoryFile()
intCo_unbounded = (gain/resetTime) * history.totalArea

# Store current total of queued unislots for next run
history.storeMcq( q.getTotalMultiQueued() * 8)

# Now we have the current mcq and the previous mcq, we can tell if a ramp up is occuring
if ((history.previousMcq < 32 ) and (history.currentMcq >= 32)):
  print 'DEBUG start of rampup detected'
  history.beginStartup()

# Try this change to stop sudden bursts
#elif (history.inStartup):
#  if (cvm >= (sp / 3) * 2):
#    history.endStartup()
if (history.inStartup):
  if (cvm >= (sp / 3) * 2):
    history.endStartup()

# Add up all the control outputs
totCo_unbounded = propCo_unbounded + intCo_unbounded
totCo = min (max(totCo_unbounded,-0.5),0.5)

realCoef = 1.0
if (history.inStartup):
  # Work out the rampup factor
  reducedCoef = options['coef'] - 1
  halfLife = options['halflife']
  timeExpired = time.time() - history.beginStartupTime
  realCoef = reducedCoef * pow(1.0/2.0,timeExpired/halfLife) + 1
  print 'DEBUG realCoef:', realCoef

# Convert controller output into a number of nodes
# to drain.

minimumTarget =  round ((totCo + 0.5) * maxToDrain) * realCoef
print 'DEBUG pcu:', propCo_unbounded,', icu:', intCo_unbounded
print 'DEBUG cvm:', cvm,', totCo:', totCo,', mt:',minimumTarget
#------------------------

drainingNodes = getDrainingNodes(nodeList)

stillDrainingNodes = putReadyNodesInService(drainingNodes,test)

# Decide what to do with this truth table (mc = multicore, sc = singlecore)
# States:
# mc jobs queued            no  yes no  yes 
# sc jobs queued            no  no  yes yes 
#
# Actions:
# start drain if nec.       no  yes no  yes
# cancel on-going drains    no  no maybe no
#
# Rationale: 
#
# If there are mc jobs queued, then start draining some systems
# if proportional controller agrees.
#
# If there are no mc jobs queued, then it's pointless to start 
# draining any systems. But if there are sc core jobs
# queued, then cancel the on-going drains to let
# the sc jobs run.
#

if (q.getTotalMultiQueued() > 0 ):
  
  if (len(stillDrainingNodes) >= minimumTarget):
    print 'Exit with no further action; already draining ',len(stillDrainingNodes)
    sys.exit()
  
  numberToDrain = minimumTarget - len(stillDrainingNodes)

  nodesToDrain = selectNodesToDrain(nodeList,numberToDrain)

  startDraining(nodesToDrain,test)
  print 'Exit having done my work'
  sys.exit()

else:

  # No mc in queue

  if (q.getTotalSingleQueued() >= 10):
    if (not options['keepgoing'] ):
      print 'Exit and cancelling because no multis and several singles queued'
      cancelSet(stillDrainingNodes,test)
      sys.exit()
    else:
      print 'Exit while no multis and several singles queued but not cancelling because keepgoing is set'
      sys.exit()
  else:
    print 'Exit no action because no multis queued and few or no singles queued'
    sys.exit()

print 'Exit but I should never reach here'
sys.exit()

